use leptos::{
    logging::log,
    prelude::{Set, Signal, Update, Write, WriteSignal},
    task::spawn_local,
};
use wasm_bindgen_futures::JsFuture;
use web_sys::{
    RtcConfiguration, RtcIceCandidate, RtcIceCandidateInit, RtcIceServer, RtcPeerConnection,
    RtcPeerConnectionIceEvent, RtcSdpType, RtcSessionDescriptionInit,
    js_sys::{Array, Reflect},
    wasm_bindgen::{JsCast, JsValue, prelude::Closure},
};

use crate::{
    signal::{
        receive_answer, receive_ice_candidate, receive_offer, send_answer, send_ice_candidate,
        send_offer,
    },
    sleep,
};

async fn local_ice_candidate_handler(username: &String, peer_connection: &RtcPeerConnection) {
    let local_ice_candidate_handler: Box<dyn Fn(String, RtcPeerConnectionIceEvent)> =
        Box::new(move |username, peer_connection_ice_event| {
            spawn_local(async move {
                {
                    let username: &String = &username;
                    async move {
                        sleep(1000).await;

                        if let Some(peer_connection_ice_candidate) =
                            peer_connection_ice_event.candidate()
                        {
                            let peer_connection_ice_candidate =
                                peer_connection_ice_candidate.as_string().unwrap();
                            send_ice_candidate(username, &peer_connection_ice_candidate)
                                .await
                                .unwrap();
                        }
                    }
                }
                .await
            });
        });

    let local_ice_candidate_handler = Closure::wrap(local_ice_candidate_handler);

    peer_connection.set_onicecandidate(Some(local_ice_candidate_handler.as_ref().unchecked_ref()));
}

async fn remote_ice_candidate_handler(username: &String, peer_connection: &RtcPeerConnection) {
    if let Ok(user_and_signal) = receive_ice_candidate(username).await {
        let peer_connection_ice_candidate_init =
            RtcIceCandidateInit::new(&user_and_signal.signal.get_data());
        let peer_connection_ice_candidate =
            RtcIceCandidate::new(&peer_connection_ice_candidate_init).unwrap();
        let peer_connection_add_ice_candidate_promise = peer_connection
            .add_ice_candidate_with_opt_rtc_ice_candidate(Some(&peer_connection_ice_candidate));
        JsFuture::from(peer_connection_add_ice_candidate_promise)
            .await
            .unwrap();
    }
}

async fn create_peer_connection_with_configuration() -> RtcPeerConnection {
    let ice_server_addresses = vec![JsValue::from("stun:stun.l.google.com:19302")]
        .into_iter()
        .collect::<Array>();
    let ice_server = RtcIceServer::new();
    ice_server.set_urls(&JsValue::from(ice_server_addresses));
    let ice_servers = vec![ice_server].into_iter().collect::<Array>();
    let rtc_configuration = RtcConfiguration::new();
    rtc_configuration.set_ice_servers(&ice_servers);
    RtcPeerConnection::new_with_configuration(&rtc_configuration).unwrap()
}

pub async fn offer(username: String) -> RtcPeerConnection {
    let peer_connection = create_peer_connection_with_configuration().await;
    let peer_connection_create_offer_promise = peer_connection.create_offer();
    let peer_connection_session_offer = JsFuture::from(peer_connection_create_offer_promise)
        .await
        .unwrap();
    log!("{:#?}", peer_connection_session_offer);
    let peer_connection_session_offer = peer_connection_session_offer
        .as_ref()
        .unchecked_ref::<RtcSessionDescriptionInit>();
    log!("{:#?}", peer_connection_session_offer);
    JsFuture::from(peer_connection.set_local_description(peer_connection_session_offer))
        .await
        .unwrap();
    let peer_connection_session_offer =
        Reflect::get(&peer_connection_session_offer, &JsValue::from_str("sdp"))
            .unwrap()
            .as_string()
            .unwrap();
    log!("{}", peer_connection_session_offer);
    let data = peer_connection_session_offer;

    if let Err(err_val) = send_offer(&username, &data).await {
        log!("Error: Send Offer | {}", err_val)
    }

    loop {
        match receive_answer(&username).await {
            Ok(received_user_and_signal_answer) => {
                log!("{:#?}", received_user_and_signal_answer);
                let peer_connection_session_answer =
                    RtcSessionDescriptionInit::new(RtcSdpType::Answer);
                peer_connection_session_answer
                    .set_sdp(received_user_and_signal_answer.signal.get_data().as_str());
                sleep(1000).await;
                JsFuture::from(
                    peer_connection.set_remote_description(&peer_connection_session_answer),
                )
                .await
                .unwrap();

                log!("{:#?}", peer_connection.connection_state());
                log!("{:#?}", peer_connection.ice_connection_state());

                local_ice_candidate_handler(&username, &peer_connection).await;
                return peer_connection;
            }
            Err(err_val) => log!("Error: Receive Answer | {}", err_val),
        }
        sleep(1000).await;
    }
}

pub async fn answer(username: String) -> RtcPeerConnection {
    loop {
        match receive_offer(&username).await {
            Ok(received_user_and_signal_offer) => {
                log!("{:#?}", received_user_and_signal_offer);
                let peer_connection = create_peer_connection_with_configuration().await;
                let peer_connection_session_offer =
                    RtcSessionDescriptionInit::new(RtcSdpType::Offer);
                peer_connection_session_offer
                    .set_sdp(received_user_and_signal_offer.signal.get_data().as_str());
                JsFuture::from(
                    peer_connection.set_remote_description(&peer_connection_session_offer),
                )
                .await
                .unwrap();
                let peer_connection_create_answer_promise = peer_connection.create_answer();
                let peer_connection_answer = JsFuture::from(peer_connection_create_answer_promise)
                    .await
                    .unwrap();
                let peer_connection_answer = peer_connection_answer
                    .as_ref()
                    .unchecked_ref::<RtcSessionDescriptionInit>();
                JsFuture::from(peer_connection.set_local_description(peer_connection_answer))
                    .await
                    .unwrap();
                let session_answer = Reflect::get(&peer_connection_answer, &JsValue::from("sdp"))
                    .unwrap()
                    .as_string()
                    .unwrap();
                log!("{}", session_answer);
                let data = session_answer;

                send_answer(&username, &data).await.unwrap();
                log!("{:#?}", peer_connection.connection_state());
                log!("{:#?}", peer_connection.ice_connection_state());

                local_ice_candidate_handler(&username, &peer_connection).await;

                return peer_connection;
            }
            Err(err_val) => log!("Error: Receive Offer | {}", err_val),
        }
        sleep(1000).await;
    }
}
